import logging
import asyncio
import meteoblue_dataset_sdk


async def myFunction():
    # query1 = {"units": {"temperature": "C", "velocity": "km/h", "length": "metric", "energy": "watts"}, "geometry": {"type": "Polygon", "coordinates": [[[2.96894, 46.041886], [2.96894, 48.216537], [10.989692, 48.216537], [10.989692, 46.041886], [2.96894, 46.041886]]]}, "format": "json", "timeIntervals": [
    #    "2017-01-01T+00:00/2019-01-31T+00:00"], "timeIntervalsAlignment": "none", "queries": [{"domain": "NEMSGLOBAL", "gapFillDomain": None, "timeResolution": "hourly", "codes": [{"code": 11, "level": "2 m above gnd"}], "transformations": [{"type": "aggregateTimeInterval", "aggregation": "mean"}, {"type": "spatialTotalAggregate", "aggregation": "mean"}]}]}

    query = {
        "units": {
            "temperature": "C",
            "velocity": "km/h",
            "length": "metric",
            "energy": "watts",
        },
        "geometry": {
            "type": "MultiPoint",
            "coordinates": [[7.57327, 47.558399, 279]],
            "locationNames": ["Basel"],
        },
        "format": "json",
        "timeIntervals": ["2019-01-01T+00:00/2019-01-01T+00:00"],
        "timeIntervalsAlignment": "none",
        "queries": [
            {
                "domain": "NEMSGLOBAL",
                "gapFillDomain": None,
                "timeResolution": "hourly",
                "codes": [{"code": 11, "level": "2 m above gnd"}],
            }
        ],
    }
    client = meteoblue_dataset_sdk.Client(apikey="xxxxxx")  # ask for key
    result = await client.query(query)
    # result is a structured object containing timestamps and data

    timeInterval = result.geometries[0].timeIntervals[0]
    data = result.geometries[0].codes[0].timeIntervals[0].data

    print(timeInterval)
    # start: 1546300800
    # end: 1546387200
    # stride: 3600

    print(data)
    # [2.89, 2.69, 2.549999, 2.3800001,

    # query1 = asyncio.create_task(mb.query(qparams))
    # query2 = asyncio.create_task(mb.query(qparams))
    # res1 = await query1
    # res2 = await query2
    # async with mb.query(query1) as response:
    #    json = await response.json()
    #    print(json)
    # print(res1)
    # print(res2)


logging.basicConfig(level=logging.DEBUG)

asyncio.run(myFunction())



from abc import ABC, abstractmethod


class AbstractCache(ABC):
    """
    Abstract class for the Cache implementation.
    Provide common methods to be implemented.
    """

    @abstractmethod
    async def set(self, key: str, value: bytes) -> None:
        """
        Store the value in the cache
        :param key: Key used to store the value
        :param value: The request data to store
        :return: None
        """
        raise NotImplementedError("Method must be implemented")

    @abstractmethod
    async def get(self, key: str) -> bytes:
        """
        Retrieve a value from the cache
        :param key: Key used to store the query to retrieve.
        :return: None if not found or the content of the file
        """
        raise NotImplementedError("Method must be implemented")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: dataset.proto
# Protobuf Python Version: 5.27.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    0,
    '',
    'dataset.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rdataset.proto\"\xdc\x05\n\x12\x44\x61tasetApiProtobuf\x12\x30\n\ngeometries\x18\x01 \x03(\x0b\x32\x1c.DatasetApiProtobuf.Geometry\x1a\xed\x01\n\x08Geometry\x12\x0e\n\x06\x64omain\x18\x01 \x01(\t\x12\x0c\n\x04lats\x18\x02 \x03(\x02\x12\x0c\n\x04lons\x18\x03 \x03(\x02\x12\x0c\n\x04\x61sls\x18\x04 \x03(\x02\x12\x15\n\rlocationNames\x18\x05 \x03(\t\x12\n\n\x02nx\x18\x06 \x01(\x12\x12\n\n\x02ny\x18\x07 \x01(\x12\x12\x16\n\x0etimeResolution\x18\x08 \x01(\t\x12\x37\n\rtimeIntervals\x18\t \x03(\x0b\x32 .DatasetApiProtobuf.TimeInterval\x12\'\n\x05\x63odes\x18\n \x03(\x0b\x32\x18.DatasetApiProtobuf.Code\x1aO\n\x0cTimeInterval\x12\x13\n\x0btimestrings\x18\x01 \x03(\t\x12\r\n\x05start\x18\x02 \x01(\x12\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x12\x12\x0e\n\x06stride\x18\x04 \x01(\x03\x1a\xa6\x02\n\x04\x43ode\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x03\x12\r\n\x05level\x18\x02 \x01(\t\x12\x10\n\x08variable\x18\x03 \x01(\t\x12\x0c\n\x04unit\x18\x04 \x01(\t\x12\x13\n\x0b\x61ggregation\x18\x05 \x01(\t\x12/\n\rtimeIntervals\x18\x06 \x03(\x0b\x32\x18.DatasetApiProtobuf.Time\x12\x0f\n\x07gddBase\x18\x07 \x01(\x02\x12\x10\n\x08gddLimit\x18\x08 \x01(\x02\x12\r\n\x05slope\x18\t \x01(\x02\x12\x0e\n\x06\x66\x61\x63ing\x18\n \x01(\x02\x12\x0e\n\x06kwPeak\x18\x0b \x01(\x02\x12\x12\n\nstartDepth\x18\x0c \x01(\x03\x12\x10\n\x08\x65ndDepth\x18\r \x01(\x03\x12\x12\n\nefficiency\x18\x0e \x01(\x02\x12\x0f\n\x07tracker\x18\x0f \x01(\x03\x1a*\n\x04Time\x12\x14\n\x0cgapFillRatio\x18\x01 \x01(\x02\x12\x0c\n\x04\x64\x61ta\x18\x02 \x03(\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dataset_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DATASETAPIPROTOBUF']._serialized_start=18
  _globals['_DATASETAPIPROTOBUF']._serialized_end=750
  _globals['_DATASETAPIPROTOBUF_GEOMETRY']._serialized_start=91
  _globals['_DATASETAPIPROTOBUF_GEOMETRY']._serialized_end=328
  _globals['_DATASETAPIPROTOBUF_TIMEINTERVAL']._serialized_start=330
  _globals['_DATASETAPIPROTOBUF_TIMEINTERVAL']._serialized_end=409
  _globals['_DATASETAPIPROTOBUF_CODE']._serialized_start=412
  _globals['_DATASETAPIPROTOBUF_CODE']._serialized_end=706
  _globals['_DATASETAPIPROTOBUF_TIME']._serialized_start=708
  _globals['_DATASETAPIPROTOBUF_TIME']._serialized_end=750
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: measurements.proto
# Protobuf Python Version: 5.27.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    0,
    '',
    'measurements.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12measurements.proto\"\x85\x04\n\x16MeasurementApiProtobuf\x12/\n\x07\x63olumns\x18\x01 \x03(\x0b\x32\x1e.MeasurementApiProtobuf.Column\x12\x12\n\nrows_count\x18\x02 \x01(\r\x12\x14\n\x0c\x63urrent_page\x18\x03 \x01(\r\x12\x15\n\rrows_per_page\x18\x04 \x01(\r\x1aH\n\x06\x43olumn\x12\x0e\n\x06\x63olumn\x18\x01 \x01(\t\x12.\n\x06values\x18\x02 \x01(\x0b\x32\x1e.MeasurementApiProtobuf.Values\x1a\xc5\x01\n\x06Values\x12\x39\n\x07strings\x18\x01 \x01(\x0b\x32&.MeasurementApiProtobuf.RepeatedStringH\x00\x12\x37\n\x06\x66loats\x18\x02 \x01(\x0b\x32%.MeasurementApiProtobuf.RepeatedFloatH\x00\x12\x37\n\x06ints64\x18\x03 \x01(\x0b\x32%.MeasurementApiProtobuf.RepeatedInt64H\x00\x42\x0e\n\x0coneof_values\x1a\x1f\n\x0eRepeatedString\x12\r\n\x05\x61rray\x18\x01 \x03(\t\x1a\"\n\rRepeatedFloat\x12\x11\n\x05\x61rray\x18\x01 \x03(\x02\x42\x02\x10\x01\x1a\"\n\rRepeatedInt64\x12\x11\n\x05\x61rray\x18\x01 \x03(\x03\x42\x02\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'measurements_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDFLOAT'].fields_by_name['array']._loaded_options = None
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDFLOAT'].fields_by_name['array']._serialized_options = b'\020\001'
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDINT64'].fields_by_name['array']._loaded_options = None
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDINT64'].fields_by_name['array']._serialized_options = b'\020\001'
  _globals['_MEASUREMENTAPIPROTOBUF']._serialized_start=23
  _globals['_MEASUREMENTAPIPROTOBUF']._serialized_end=540
  _globals['_MEASUREMENTAPIPROTOBUF_COLUMN']._serialized_start=163
  _globals['_MEASUREMENTAPIPROTOBUF_COLUMN']._serialized_end=235
  _globals['_MEASUREMENTAPIPROTOBUF_VALUES']._serialized_start=238
  _globals['_MEASUREMENTAPIPROTOBUF_VALUES']._serialized_end=435
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDSTRING']._serialized_start=437
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDSTRING']._serialized_end=468
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDFLOAT']._serialized_start=470
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDFLOAT']._serialized_end=504
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDINT64']._serialized_start=506
  _globals['_MEASUREMENTAPIPROTOBUF_REPEATEDINT64']._serialized_end=540
# @@protoc_insertion_point(module_scope)



from meteoblue_dataset_sdk.caching import FileCache
from meteoblue_dataset_sdk.client import Client

import shutil
import tempfile
import zlib
from pathlib import Path

from freezegun import freeze_time
from later.unittest import mock

# until we upgrade to >=3.8
from later.unittest.backport.async_case import IsolatedAsyncioTestCase


class TestFileCache(IsolatedAsyncioTestCase):
    def setUp(self):
        self.params = {
            "units": {
                "temperature": "C",
                "velocity": "km/h",
                "length": "metric",
                "energy": "watts",
            },
            "geometry": {
                "type": "MultiPoint",
                "coordinates": [[7.57327, 47.558399, 279]],
                "locationNames": ["Basel"],
            },
            "format": "json",
            "timeIntervals": ["2019-01-01T+00:00/2019-01-01T+00:00"],
            "timeIntervalsAlignment": "none",
            "queries": [
                {
                    "domain": "NEMSGLOBAL",
                    "gapFillDomain": None,
                    "timeResolution": "hourly",
                    "codes": [{"code": 11, "level": "2 m above gnd"}],
                }
            ],
        }

        self.cache_path = Path(tempfile.gettempdir(), "mb_cache")
        if not self.cache_path.exists():
            self.cache_path.mkdir()
        self.dir_hash = "f4e"
        self.file_hash = "bb32678055481aa33398a8a7afaa5"
        self.dir_path = Path(tempfile.gettempdir(), "mb_cache", self.dir_hash)
        self.file_path = Path(
            tempfile.gettempdir(), "mb_cache", self.dir_hash, self.file_hash
        )

    def test__hash_to_paths(self):
        self.assertIsNone(FileCache()._hash_to_path(""))
        self.assertEqual(
            FileCache()._hash_to_path("88bac95f31528d13a072c05f2a1cf371"),
            Path(
                tempfile.gettempdir(),
                "mb_cache",
                "88b",
                "ac95f31528d13a072c05f2a1cf371",
            ),
        )

    def test_path(self):
        file_cache = FileCache()
        self.assertEqual(file_cache.cache_path, Path(tempfile.gettempdir(), "mb_cache"))
        file_cache = FileCache(path="/tmp/test_cache")
        self.assertEqual("/tmp/test_cache", file_cache.cache_path.as_posix())

    @mock.patch("pathlib.Path.exists", return_value=True)
    @mock.patch("aiofiles.os.stat")
    async def test__is_cached_file_valid(self, mock_stat, mock_path):
        file_cache = FileCache()
        with freeze_time("2020-01-01 12:00:00"):
            mock_stat.return_value.st_mtime = freeze_time(
                "2020-01-01 11:59:00"
            ).time_to_freeze.timestamp()
            self.assertTrue(await file_cache._is_cached_file_valid(Path("somepath")))

        with freeze_time("2020-01-01 12:00:00"):
            mock_stat.return_value.st_mtime = freeze_time(
                "2020-01-01 06:00:00"
            ).time_to_freeze.timestamp()
            self.assertFalse(await file_cache._is_cached_file_valid(Path("somepath")))

    @mock.patch("meteoblue_dataset_sdk.caching.FileCache._is_cached_file_valid")
    async def test_get(self, mock__is_cached_file_valid):

        self.dir_path.mkdir()
        with open(self.file_path, "wb") as file:
            file.write(zlib.compress(bytes('{"response": "data"}', "utf-8")))

        mock__is_cached_file_valid.return_value = True
        file_cache = FileCache()
        key = Client(apikey="124")._hash_params(self.params)
        self.assertEqual(await file_cache.get(key), b'{"response": "data"}')
        mock__is_cached_file_valid.assert_called_with(
            Path(tempfile.gettempdir(), "mb_cache", self.dir_hash, self.file_hash)
        )

    @mock.patch("meteoblue_dataset_sdk.caching.FileCache._is_cached_file_valid")
    async def test_set(self, mock__is_cached_file_valid):
        mock__is_cached_file_valid.return_value = False
        file_cache = FileCache()
        key = Client(apikey="124")._hash_params(self.params)
        await file_cache.set(key, bytes('{"someData": "superData"}', "utf-8"))
        with open(self.file_path, "rb") as file:
            self.assertEqual(zlib.decompress(file.read()), b'{"someData": "superData"}')

    def tearDown(self):
        if self.cache_path.exists():
            shutil.rmtree(self.cache_path, ignore_errors=True)
        super(TestFileCache, self).tearDown()



from meteoblue_dataset_sdk import ApiError, Client

import asyncio
import logging
import os
import unittest


class TestQuery(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)

    def test_invalid_query(self):
        client = Client("invalid_api_key")
        with self.assertRaises(ApiError):
            result = asyncio.run(client.query({"invalid": "query"}))
            self.assertEqual(
                result, "API returned error message: Value required for key 'geometry'."
            )

    def test_simple_query(self):
        query = {
            "units": {
                "temperature": "C",
                "velocity": "km/h",
                "length": "metric",
                "energy": "watts",
            },
            "geometry": {
                "type": "MultiPoint",
                "coordinates": [[7.57327, 47.558399, 279]],
                "locationNames": ["Basel"],
            },
            "format": "json",
            "timeIntervals": ["2019-01-01T+00:00/2019-01-01T+00:00"],
            "timeIntervalsAlignment": "none",
            "runOnJobQueue": True,
            "queries": [
                {
                    "domain": "NEMSGLOBAL",
                    "gapFillDomain": None,
                    "timeResolution": "hourly",
                    "codes": [{"code": 11, "level": "2 m above gnd"}],
                }
            ],
        }

        client = Client(os.environ["APIKEY"])
        result = asyncio.run(client.query(query))
        geo = result.geometries[0]
        timeInterval = result.geometries[0].timeIntervals[0]
        variable = geo.codes[0]
        data = variable.timeIntervals[0].data

        self.assertEqual(geo.domain, "NEMSGLOBAL")
        self.assertEqual(geo.lats, [47.66651916503906])
        self.assertEqual(geo.lons, [7.5])
        self.assertEqual(geo.asls, [499.7736511230469])
        self.assertEqual(geo.locationNames, ["Basel"])
        self.assertEqual(geo.nx, 1)
        self.assertEqual(geo.ny, 1)
        self.assertEqual(geo.timeResolution, "hourly")

        self.assertEqual(variable.code, 11)
        self.assertEqual(variable.level, "2 m above gnd")
        self.assertEqual(variable.unit, "°C")
        self.assertEqual(variable.aggregation, "none")

        self.assertEqual(timeInterval.start, 1546300800)
        self.assertEqual(timeInterval.end, 1546383600 + 3600)
        self.assertEqual(timeInterval.stride, 3600)

        self.assertEqual(
            data,
            [
                2.890000104904175,
                2.690000057220459,
                2.549999952316284,
                2.380000114440918,
                2.2699999809265137,
                2.119999885559082,
                1.9900000095367432,
                1.8300000429153442,
                1.8200000524520874,
                2.0999999046325684,
                2.430000066757202,
                2.9200000762939453,
                3.7200000286102295,
                3.930000066757202,
                3.9100000858306885,
                3.5299999713897705,
                3.130000114440918,
                2.880000114440918,
                2.6500000953674316,
                2.4600000381469727,
                2.2799999713897705,
                2.0299999713897705,
                1.690000057220459,
                1.3799999952316284,
            ],
        )

    def test_complex_query(self):
        # This query is read half of Europe and 2 years of data
        # The API will refuse to run it directly
        query_complex = {
            "units": {
                "temperature": "C",
                "velocity": "km/h",
                "length": "metric",
                "energy": "watts",
            },
            "geometry": {
                "type": "Polygon",
                "coordinates": [
                    [
                        [2.96894, 46.041886],
                        [2.96894, 48.216537],
                        [10.989692, 48.216537],
                        [10.989692, 46.041886],
                        [2.96894, 46.041886],
                    ]
                ],
            },
            "format": "json",
            "timeIntervals": ["2017-01-01T+00:00/2019-01-31T+00:00"],
            "timeIntervalsAlignment": "none",
            "queries": [
                {
                    "domain": "NEMSGLOBAL",
                    "gapFillDomain": None,
                    "timeResolution": "hourly",
                    "codes": [{"code": 11, "level": "2 m above gnd"}],
                    "transformations": [
                        {"type": "aggregateTimeInterval", "aggregation": "mean"},
                        {"type": "spatialTotalAggregate", "aggregation": "mean"},
                    ],
                }
            ],
        }

        client = Client(os.environ["APIKEY"])
        result = asyncio.run(client.query(query_complex))
        geo = result.geometries[0]
        timestamps = geo.timeIntervals[0].timestrings
        variable = geo.codes[0]
        data = variable.timeIntervals[0].data

        self.assertEqual(geo.domain, "NEMSGLOBAL")
        self.assertEqual(len(geo.lats), 1)
        self.assertEqual(len(geo.lons), 1)
        self.assertAlmostEqual(geo.lats[0], 47.12916946411133, 3)
        self.assertAlmostEqual(geo.lons[0], 6.97930908203125, 3)
        self.assertEqual(geo.nx, 1)
        self.assertEqual(geo.ny, 1)
        self.assertEqual(geo.timeResolution, "total")

        self.assertEqual(variable.code, 11)
        self.assertEqual(variable.level, "2 m above gnd")
        self.assertEqual(variable.unit, "°C")
        self.assertEqual(variable.aggregation, "mean")

        self.assertEqual(timestamps, ["20170101T0000-20190131T235959"])
        self.assertEqual(len(data), 1)
        self.assertAlmostEqual(data[0], 8.519842147827148, 3)



""""
meteoblue dataset client
"""

import asyncio
import copy
import hashlib
import json
import logging
from contextlib import asynccontextmanager

import aiohttp

from .protobuf.dataset_pb2 import DatasetApiProtobuf
from .protobuf.measurements_pb2 import MeasurementApiProtobuf
from .utils import run_async


class ClientConfig(object):
    def __init__(self, apikey: str):
        # urls
        # following job id
        self.status_url = "http://my.meteoblue.com/queue/status/{}"
        # following api key
        self.query_url = "http://my.meteoblue.com/dataset/query?apikey={}"
        # following job id
        self.result_url = "http://queueresults.meteoblue.com/{}"

        # http
        self.http_max_retry_count = 5
        self.http_retry_sleep_duration = 1

        # other config
        self.api_key = apikey
        self.queue_retry_sleep_duration = 5


class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class ApiError(Error):
    """Exception raised for errors by the dataset API

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class Client(object):
    def __init__(self, apikey: str, cache=None):
        self._config = ClientConfig(apikey)
        self.cache = cache

    @asynccontextmanager
    async def _fetch(
        self,
        session: aiohttp.ClientSession,
        method: str,
        url: str,
        body_dict: dict = None,
        query_params: dict = None,
    ):
        """
        Fetch data from an URL and try for error 5xx or timeouts.
        Codes other than 2xx will throw an exception.
        :param session: an active aiohttp.ClientSession
        :param method: HTTP verb to use for the request
        :param url: url to fetch data from
        :param body_dict: parameters transferred in the body
        :param query_params: parameters transferred as query parameters in the url
        :return: ClientResponse object from aiohttp lib
        """
        logging.debug(f"Getting url {method} {url}")
        for retry in range(self._config.http_max_retry_count):
            async with session.request(
                method, url, json=body_dict, params=query_params
            ) as response:
                # return if successful
                if 200 <= response.status <= 299:
                    yield response
                    return

                # meteoblue APIs return a JSON encoded error message
                if response.status == 400 or response.status == 404 or response.status == 403 or response.status == 422 or response.status == 500:
                    json_response = await response.json()
                    # TODO: dataset api returns object with 'error_message'
                    # measurement api object with 'reason'
                    error_message = ""
                    if "error_message" in json_response:
                        error_message = json_response["error_message"]
                    else:
                        error_message = json_response["reason"]
                    logging.debug(f"API returned error message: {error_message}")
                    raise ApiError(error_message)

                if retry == self._config.http_max_retry_count - 1:
                    logging.error(f"API returned unexpected error: {response.content}")
                    raise Exception("API returned unexpected error", response.content)

                # retry mechanism
                # check if request timed out or backend threw an error
                # if response.status == 408 or 500 <= response.status <= 599:
                logging.info(f"Request failed or timed out. Try {retry}")
                logging.debug(response)

    @asynccontextmanager
    async def _run_on_job_queue(self, session: aiohttp.ClientSession, params: dict):
        """
        Run the query on a job queue
        :param session: ClientSession
        :param params: query parameters for meteoblue dataset api
        :return: ClientResponse object from aiohttp lib
        """

        # Start job on a job queue
        logging.info("Starting job on queue")
        params["runOnJobQueue"] = True
        url = self._config.query_url.format(self._config.api_key)
        async with self._fetch(session, "POST", url, body_dict=params) as response:
            response_json = await response.json()

        # Wait until the job is finished
        job_id = response_json["id"]
        logging.info(f"Waiting until job has finished (job id {job_id})")
        status_url = self._config.status_url.format(job_id)
        while True:
            async with self._fetch(session, "GET", status_url) as response:
                json_data = await response.json()
            status = json_data["status"]
            logging.debug(f"Job status is {status}")
            if status == "finished":
                break
            if status == "deleted":
                raise ApiError("Job was canceled")
            if status == "error":
                raise ApiError(json_data["error_message"])
            logging.info(
                f"Waiting 5 seconds for job to complete. Status: {status}, \
                job id {job_id}"
            )
            await asyncio.sleep(self._config.queue_retry_sleep_duration)

        # Fetch the job queue result
        result_url = self._config.result_url.format(job_id)
        logging.debug(f"Fetching result for job id {job_id}")
        async with self._fetch(session, "GET", result_url) as response:
            yield response

    @asynccontextmanager
    async def _query_raw(self, params: dict):
        """
        Query meteoblue dataset api asynchronously and return a ClientResponse object
        using context manager
        :param params: query parameters
            see https://docs.meteoblue.com/en/apis/environmental-data/dataset-api
        :return: ClientResponse object from aiohttp lib
        """

        # always try to execute without job queue first:
        params["runOnJobQueue"] = False
        async with aiohttp.ClientSession() as session:
            # Try to run the job directly
            # In case the API throws an error, try to run it on a job queue
            try:
                url = self._config.query_url.format(self._config.api_key)
                async with self._fetch(
                    session, "POST", url, body_dict=params
                ) as response:
                    yield response
            except ApiError as error:
                # Run on a job queue in case the api throws this error
                if error.message != "This job must be executed on a job-queue":
                    raise
                async with self._run_on_job_queue(session, params) as response:
                    yield response

    @staticmethod
    def _hash_params(params: dict) -> str:
        return hashlib.md5(json.dumps(params).encode()).hexdigest()

    @staticmethod
    def _parse_dataset(data):
        msg = DatasetApiProtobuf()
        msg.ParseFromString(data)
        return msg

    async def query(self, params: dict):
        """
        Query meteoblue dataset api asynchronously, transfer data using protobuf and
        return a structured object

        :param params:
            query parameters,
            see https://docs.meteoblue.com/en/apis/environmental-data/dataset-api
        :return: DatasetApiProtobuf object
        """
        # copy params object before making changes to it
        params = copy.copy(params)
        params["format"] = "protobuf"
        cache_key = ""
        if self.cache:
            cache_key = self._hash_params(params)
            cached_query_results = await self.cache.get(cache_key)
            if cached_query_results:
                return self._parse_dataset(cached_query_results)

        async with self._query_raw(params) as response:
            data = await response.read()
            if self.cache:
                await self.cache.set(cache_key, data)
            return self._parse_dataset(data)

    def querySync(self, params: dict):
        """
        Query meteoblue dataset api synchronously for sequential usage.
        Prefer query_sync in order to respect python semantic.
        :param params:
             query parameters.
             see https://docs.meteoblue.com/en/apis/environmental-data/dataset-api
        :return: DatasetApiProtobuf object
        """
        return run_async(self.query, params)

    def query_sync(self, params: dict):
        """
        Exactly the same as querySync but using underscore in the name.
        Keeping querySync for backward compatibility.
        """
        return self.querySync(params)

    def measurement_sync(self, path: str, params: dict):
        """
        Query meteoblue measurement api synchronously for sequential usage.

        :param path: path of request
        :param params: query parameters

        :return: MeasurementApiProtobuf object
        """
        return run_async(self.measurement_query, path, params)

    async def measurement_query(self, path: str, params: dict):
        """
        Query meteoblue measurement api asynchronously, transfer data using protobuf and
        return a structured object

        :param path: path of request
        :param params: query parameters

        :return: MeasurementApiProtobuf object
        """
        # copy params object before making changes to it
        params = copy.copy(params)
        params["format"] = "protobuf"
        cache_key = ""
        if self.cache:
            cache_key = self._hash_params_measurements(path, params)
            cached_query_results = await self.cache.get(cache_key)
            if cached_query_results:
                self._parse_measurements(cached_query_results)

        async with self._query_measurement_api(path, params) as response:
            data = await response.read()
            if self.cache:
                await self.cache.set(cache_key, data)
            return self._parse_measurements(data)

    @staticmethod
    def _parse_measurements(data):
        msg = MeasurementApiProtobuf()
        msg.ParseFromString(data)
        return msg

    @staticmethod
    def _hash_params_measurements(path: str, params: dict) -> str:
        return hashlib.md5(path.encode() + json.dumps(params).encode()).hexdigest()

    @asynccontextmanager
    async def _query_measurement_api(self, path: str, params: dict):
        """
        Query meteoblue measurement api asynchronously and return a
        ClientResponse object using context manager

        :param path: path of request
        :param params: query parameters
        :return: ClientResponse object from aiohttp lib
        """
        base_url = "https://measurements-api.meteoblue.com"
        params["apikey"] = self._config.api_key
        url = base_url + path
        async with aiohttp.ClientSession() as session:
            async with self._fetch(
                session=session,
                method="GET",
                url=url,
                body_dict=None,
                query_params=params,
            ) as response:
                yield response



from math import isnan
from meteoblue_dataset_sdk import ApiError, Client
from meteoblue_dataset_sdk.caching import FileCache
from meteoblue_dataset_sdk.protobuf.measurements_pb2 import MeasurementApiProtobuf

import asyncio
import logging
import os
import unittest


class TestMeasurementQuery(unittest.TestCase):
    def assertIsNaN(self, value: float):
        self.assertTrue(isnan(value))

    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)

    @staticmethod
    def getColumnValues(message: MeasurementApiProtobuf, column: str):
        filtered_column = [col for col in message.columns if col.column == column]
        if filtered_column:
            col = filtered_column[0]
            col_type = col.values.WhichOneof("oneof_values")
            if col_type == "strings":
                return col.values.strings.array
            elif col_type == "floats":
                return col.values.floats.array
            elif col_type == "ints64":
                return col.values.ints64.array
            elif col_type == "ints32":
                return col.values.ints32.array
            elif col_type == "uints64":
                return col.values.uint64.array
            elif col_type == "uints32":
                return col.values.uints32.array
            elif col_type == "bools":
                return col.values.bools.array

    def test_invalid_table_key(self):
        client = Client(os.environ["APIKEY"])
        path = "/v2/dwdClimateHourly/raw/invalid/get"
        expected_error_message = (
            "API returned error message:"
            " Unknown table invalid for provider dwdClimateHourly"
        )
        with self.assertRaises(ApiError):
            result = asyncio.run(client.measurement_query(path, {"invalid": "query"}))
            self.assertEqual(
                result,
                expected_error_message,
            )

    def test_invalid_api_key(self):
        client = Client("invalid_api_key")
        provider = "dwdClimateHourly"
        table = "measurement"
        path = f"/v2/{provider}/raw/{table}/get"

        with self.assertRaises(ApiError):
            result = asyncio.run(client.measurement_query(path, {"invalid": "query"}))
            self.assertEqual(result, "Apikey was specified, but is not valid.")

    def test_simple_query(self):
        query = {
            "timeStart": "2024-05-01T12:00:00",
            "timeEnd": "2024-05-02T12:00:00",
            "limit": 10,
            "sort": "asc",
            "stations": ["00044"],
            "fields": [
                "id",
                "timestamp",
                "lat",
                "lon",
                "asl",
                "temperature_2mAbvGnd_atTimestamp_none_degCels",
            ],
        }
        provider = "dwdClimate10Minute"
        table = "measurement"
        path = f"/v2/{provider}/raw/{table}/get"

        cache = FileCache()
        client = Client(os.environ["APIKEY"], cache=cache)
        result = asyncio.run(client.measurement_query(path, query))
        rows_per_page = result.rows_per_page
        current_page = result.current_page
        rows_count = result.rows_count
        timestamps = self.getColumnValues(result, "timestamp")
        station_ids = self.getColumnValues(result, "id")
        lats = self.getColumnValues(result, "lat")
        lons = self.getColumnValues(result, "lon")
        asls = self.getColumnValues(result, "asl")
        temperatures = self.getColumnValues(
            result, "temperature_2mAbvGnd_atTimestamp_none_degCels"
        )

        self.assertEqual(rows_per_page, 10)
        self.assertEqual(current_page, 1)
        self.assertEqual(rows_count, 10)
        self.assertEqual(
            timestamps,
            [
                1714564800,
                1714565400,
                1714566000,
                1714566600,
                1714567200,
                1714567800,
                1714568400,
                1714569000,
                1714569600,
                1714570200,
            ],
        )
        self.assertEqual(
            station_ids,
            [
                "00044",
                "00044",
                "00044",
                "00044",
                "00044",
                "00044",
                "00044",
                "00044",
                "00044",
                "00044",
            ],
        )
        self.assertEqual(
            lats,
            [
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
                52.93360137939453,
            ],
        )
        self.assertEqual(
            lons,
            [
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
                8.237000465393066,
            ],
        )
        self.assertEqual(
            asls, [44.0, 44.0, 44.0, 44.0, 44.0, 44.0, 44.0, 44.0, 44.0, 44.0]
        )
        self.assertIsNaN(temperatures[0])
        self.assertAlmostEqual(temperatures[1], 25.6, 1)
        self.assertAlmostEqual(temperatures[2], 26, 1)
        self.assertIsNaN(temperatures[3])
        self.assertAlmostEqual(temperatures[4], 26.1, 1)
        self.assertAlmostEqual(temperatures[5], 26.1, 1)
        self.assertIsNaN(temperatures[6])
        self.assertAlmostEqual(temperatures[7], 26.4, 1)
        self.assertAlmostEqual(temperatures[8], 26.1, 1)
        self.assertIsNaN(temperatures[9])



import asyncio
import threading


def run_async(func, *args, **kwargs):
    """
    In case a jupyter notebook is running which uses a eventloop, which would raise
    "asyncio.run() cannot be called from a running event loop"
    move the asyncio execution to another thread.
    See https://stackoverflow.com/q/55409641
    """
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None
    if loop and loop.is_running() and is_jupyter_notebook():
        thread = RunThread(func, args, kwargs)
        thread.start()
        thread.join()
        if thread.exception:
            raise thread.exception
        return thread.result
    else:
        return asyncio.run(func(*args, **kwargs))


class RunThread(threading.Thread):
    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.exception = None
        super().__init__()

    def run(self):
        try:
            self.result = asyncio.run(self.func(*self.args, **self.kwargs))
        except Exception as e:
            self.exception = e


def is_jupyter_notebook():
    """
    Return true if running in a python notebook
    See https://stackoverflow.com/a/39662359
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "SpyderShell":
            return True  # Spyder IDE
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter



import nox

TARGETS = ["meteoblue_dataset_sdk", "tests", "noxfile.py"]


@nox.session(reuse_venv=True)
def lint(session):
    lint_tools = ["flake8", "isort"]
    session.install(*lint_tools)
    session.run("isort", *TARGETS)
    session.run("flake8", *TARGETS)


@nox.session(reuse_venv=True)
def lint_dev(session):
    lint_tools = ["flake8", "black", "isort"]
    session.install(*lint_tools)
    session.run("isort", "-f", *TARGETS)
    session.run("black", *TARGETS)
    session.run("flake8", *TARGETS)


@nox.session(reuse_venv=True)
def tests(session):
    session.install("-e", ".[dev]")
    session.run("pytest", "-v")



import datetime
import logging
import tempfile
import zlib
from pathlib import Path
from typing import Optional

import aiofiles
import aiofiles.os

from .abstractcache import AbstractCache

DEFAULT_CACHE_DIR = "mb_cache"
# 7200s == 2h
DEFAULT_CACHE_DURATION = 7200


class FileCache(AbstractCache):
    def __init__(
        self,
        path: Optional[str] = None,
        max_age: int = DEFAULT_CACHE_DURATION,
        compression_level: int = zlib.Z_DEFAULT_COMPRESSION,
    ):
        """
        Local file storage class
        :param path: Custom local cache storage path. Default to
        /SYSTEM_TEMP_FOLDER/mb_cache
        :param max_age: Cache retention period in seconds
        :param compression_level: Zlib compression level used to compressed the bytes
        input
        """
        cache_path = Path(tempfile.gettempdir(), DEFAULT_CACHE_DIR)
        if path is not None:
            cache_path = Path(path)
        if not cache_path.exists():
            cache_path.mkdir(parents=True)

        self.cache_path = cache_path
        self.max_age = max_age
        self.compression_level = compression_level

    async def set(self, key: str, value: bytes) -> None:
        """
        Hash the query params and use its value a directory and file name.
        If there is already a valid cache file existing, it will exit. Otherwise it
        will write to a temporary file before renaming it.
        :param key: Key used to store the value
        :param value: The request data to store
        """
        if not key:
            return
        cache_file_path = self._hash_to_path(key)

        try:
            await aiofiles.os.stat(cache_file_path.parent)
        except FileNotFoundError:
            await aiofiles.os.mkdir(cache_file_path.parent)

        if await self._is_cached_file_valid(cache_file_path):
            return
        temp_file_path = f"{cache_file_path}~"
        async with aiofiles.open(temp_file_path, "wb") as file:
            await file.write(zlib.compress(value, self.compression_level))
        await aiofiles.os.rename(temp_file_path, cache_file_path)

    async def get(self, key: str) -> Optional[bytes]:
        if not key:
            return
        file_path = self._hash_to_path(key)

        if not await self._is_cached_file_valid(file_path):
            return
        try:
            async with aiofiles.open(file_path, "rb") as f:
                return zlib.decompress(await f.read())
        except (OSError, IOError) as e:
            logging.error(f"error while reading the file {file_path}", e)
            return

    async def _is_cached_file_valid(self, file_path: Path) -> bool:
        if not file_path.exists():
            return False
        stats = await aiofiles.os.stat(file_path)
        file_modification_timestamp = int(stats.st_mtime)
        ts_as_datetime = datetime.datetime.fromtimestamp(file_modification_timestamp)
        cache_duration = datetime.datetime.now() - ts_as_datetime
        return cache_duration.seconds < self.max_age

    def _hash_to_path(self, key_hash: str) -> Optional[Path]:
        """
        Split the hash in two part: the directory and the file.
        :param key_hash: Request parameters to use a key.
        :return: The first 3 characters of the hash as the directory name and the rest
        as the filename.
        """
        if not key_hash:
            return
        return Path(self.cache_path, key_hash[:3], key_hash[3:])
